
/*
 *  backtrace dump tool
 *  Copyright (C) 2001 xiaobing tu <xiaobing.tu@intel.com>
 *
 *  GPLv2
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <sys/ptrace.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <sys/types.h>
#include <unistd.h>
#include <dirent.h>
#include <elf.h>
#include <stdbool.h>
#include <errno.h>
#include <signal.h>
#include <pthread.h>
#include <stdarg.h>
#include <fcntl.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <string.h>


#include "symbols.h"



typedef struct stack_info {
	int pid;
	int ppid;//parent pid
	char p_name[NAME_LENGTH]; // thread/process name
	int stack_depth;
	int is_thread;
	int child_count;
	int p_index;
	unsigned long ebp[MAX_STACK_LENGTH];//user stack for this process
	char kernel_stack[MAX_KERNELSTACK_SIZE];
} stack_info;

stack_info StackAll[TASK_NUM];
int g_index = 0;
long read_all_stack_time = 0;
bool is64OS = false;
static inline long calc_timediff(struct timeval *t0, struct timeval *t1)
{
	return (t1->tv_sec - t0->tv_sec) * 1000 +
		(t1->tv_usec - t0->tv_usec) / 1000;
}

int judge_64_OS()
{
	system("uname -a > 64_os.txt");
	FILE *fp  = NULL;
	char data[1024];
	fp = fopen("64_os.txt","r");
	if (fp)
	{
		fgets(data,1024,fp);
		char *str = NULL;
	        str = strstr(data,"x86_64");
		if ( str )
			return 1;
		else
			return 0;
		fclose(fp);
	}	
	return 0;
}

mapinfo *parse_map_line(char *line)
{
	mapinfo *mi;
	int len = strlen(line);

	if(len < 1) 
		return 0;
	line[--len] = 0;

	if(len < 50) 
		return 0;

	if(line[20] != 'x') 
		return 0;

	mi = malloc(sizeof(mapinfo) + (len - 47));
	if(mi == NULL) 
		return 0;

	mi->start = strtoul(line, 0, 16);
	mi->end = strtoul(line + 9, 0, 16);
	mi->exidx_start = mi->exidx_end = 0;
	mi->symbols = 0;
	mi->next = 0;
	strcpy(mi->name, line + 49);

	return mi;
}

mapinfo *parse_map_line_64(char *line)
{
	mapinfo *mi;
	char * token = NULL;
	char * p = NULL;
	int len = strlen(line);
	unsigned long start_addr = 0;
	unsigned long end_addr = 0;
	if (len < 2)
		return 0;
	line[--len] = 0;
	mi = malloc(sizeof(mapinfo));
	if (mi == NULL)
		return 0;

	token = strtok(line, " ");
	char buf1[50] = {0,};

	p = strchr(token, '-');
	strncpy(buf1, token, p-token);

	char buf2[50] = {0,};
	strncpy(buf2, p, strlen(p));
	start_addr = strtoull(buf1, NULL, 16);
	end_addr  = strtoull(p+1, NULL, 16);
	//rwx
	token = strtok(NULL, " ");

	p = strchr(token, 'x');
	if (p == NULL)
		return NULL;

	//map file 
	int counter = 0;
	while((token = strtok(NULL, " "))){
		p = token;
		counter++;		
	}
	if (counter != 4)
		return NULL;

	//fill struct
	mi->start = start_addr;
	mi->end = end_addr;
	mi->exidx_start = mi->exidx_end = 0;
	mi->symbols = 0;
	mi->next = 0;
	strcpy(mi->name, p);
	return mi;
}


static void parse_elf32(mapinfo *milist, pid_t pid)
{
	mapinfo *mi;
	struct symbol_table *table = NULL;
	for (mi = milist; mi != NULL; mi = mi->next) {
		Elf32_Ehdr  ehdr;
		Elf32_Ehdr  *ehdr1;
		memset(&ehdr, 0, sizeof(Elf32_Ehdr));
		/*
		 * Read in sizeof(Elf32_Ehdr) worth of data from
		 * the beginning of mapped section.
		 */
		struct stat sb;
		int length;
		char *base;
		int fd = open(mi->name, O_RDONLY);
		if(fd < 0) {
			continue ;
		}

		fstat(fd, &sb);
		length = sb.st_size;
		base = mmap(NULL, length, PROT_READ, MAP_PRIVATE, fd, 0);
		if(!base) {

			goto out_close;
		}

		// Parse the file header
		ehdr1 = ((Elf32_Ehdr*)base);
		memcpy(&ehdr,ehdr1,sizeof(Elf32_Ehdr));
out_unmap:
		munmap(base, length);
out_close:
		close(fd);
		if (IS_ELF(ehdr)) {
			/* Try to load symbols from this file */
			mi->symbols = symbol_tables_create(mi->name);
		}
	}
}

static void parse_elf64(mapinfo *milist, pid_t pid)
{
	mapinfo *mi;
	struct symbol_table *table = NULL;
	for (mi = milist; mi != NULL; mi = mi->next) {
		Elf64_Ehdr  ehdr;
		Elf64_Ehdr  *ehdr1;
		memset(&ehdr, 0, sizeof(Elf64_Ehdr));
		/*
		 * Read in sizeof(Elf64_Ehdr) worth of data from
		 * the beginning of mapped section.
		 */
		struct stat sb;
		int length;
		char *base;
		int fd = -1;
	        fd = open(mi->name, O_RDONLY);	
		if(fd < 0) {
			continue ;
		}
		fstat(fd, &sb);
		length = sb.st_size;
		base = mmap(NULL, length, PROT_READ, MAP_PRIVATE, fd, 0);
		if(!base) {

			goto out_close;
		}
		char * p = (char *)base+18;
		bool  is_Elf32 = false;
		if (*p != 0x3e)
			is_Elf32 = true;
		// Parse the file header
		ehdr1 = ((Elf64_Ehdr*)base);
		memcpy(&ehdr,ehdr1,sizeof(Elf64_Ehdr));
		if (IS_ELF(ehdr)) {
			/* Try to load symbols from this file */
			//			LOGV("load symbols from file...\n");
			//			LOGV("symbol_tables_create64\n");
			if(is_Elf32) {
				mi->symbols = symbol_tables_create(mi->name);
			}
			else
				mi->symbols = symbol_tables_create64(mi->name);
		}

out_unmap:
		munmap(base, length);
out_close:
		close(fd);

	}
}

void parse_elf(mapinfo *milist, pid_t pid)
{
   if(judge_64_OS()) {
	   parse_elf64(milist,pid);
   }
   else  {
	   parse_elf32(milist,pid);
   }
}


static bool read_single_stack(char * data, int  index)
{
	unsigned long tmp;
	FILE* fp = NULL;
	int iCount = 0,i = 0;
	bool access_ok = false;
	char *pKernel = NULL;

	if(index >= TASK_NUM)
		return access_ok;

	sprintf(data, "%s/stack", data);
	fp = fopen(data, "r");
	if (fp) {
		pKernel = StackAll[index].kernel_stack;
		while(fgets(data, PATH_LENGTH, fp)) {
			if( strncmp("userspace", data, 9)==0 ) {
				iCount++;
			} else {
				if (iCount == 0) {
					memcpy(pKernel,data,strlen(data));
					pKernel = pKernel + strlen(data);
				}
			}

			if(iCount > 0 && i <32) {
				if (iCount !=1)	{
					tmp = strtoul(data,0,16);
					StackAll[index].ebp[i] = tmp;
					i++;
				}
				iCount++;
			}
			access_ok = true;
		}

		fclose(fp);
	}

	if(access_ok) {
		/* update the stack depth for this process*/
		StackAll[index].stack_depth = i;
	}

	return access_ok;
}

/* Return true if some thread is not detached cleanly */
static bool dump_sibling_thread_stack(unsigned pid, int parent_index)
{
	char task_path[PATH_LENGTH];
	char data[PATH_LENGTH];
	unsigned int data1[MAX_STACK_LENGTH];
	sprintf(task_path, "/proc/%d/task", pid);
    DIR *d;
    struct dirent *de;
    int need_cleanup = 0;
    d = opendir(task_path);
    /* Bail early if cannot open the task directory */
    if (d == NULL) {
        printf("Cannot open /proc/%d/task\n", pid);
        return false;
    }
    while ((de = readdir(d)) != NULL) {
        unsigned new_tid;
        /* Ignore "." and ".." */
        if (!strcmp(de->d_name, ".") || !strcmp(de->d_name, ".."))
            continue;
        new_tid = atoi(de->d_name);
        /* The main thread at fault has been handled individually */
        if (new_tid != pid)
         {
		sprintf(data, "%s/%d", task_path,new_tid);
		if ( read_single_stack(data, g_index) ) {
			StackAll[g_index].pid  = new_tid;
			StackAll[g_index].ppid = pid;
			StackAll[g_index].is_thread = 1;
			StackAll[g_index].p_index = parent_index;
			StackAll[parent_index].child_count++;
			g_index++;
		}

	}
        /* Skip this thread if cannot ptrace it */
     }
    closedir(d);
    return need_cleanup != 0;
}


int read_all_stack()
{	
	pid_t traced_process;
	long ins;
	FILE *fp;
	DIR * dp;
	struct dirent *filename;
	struct timeval currentTime, stillafstarttime;
	char data[PATH_LENGTH];
	int stack_depth = 0;
	int index_tmp = 0;

	gettimeofday(&stillafstarttime,0);
	printf("parse_stack \n");

	dp = opendir("/proc/");

	if (!dp)
	{
		printf("open /proc directory error\n");
		return 0;
	}

	while ((filename=readdir(dp)))
	{
		if( filename->d_name[0] > '0' &&  filename->d_name[0] < '9' )
		{
			traced_process = atoi(filename->d_name);
			/* Clear stack pointer records */
			StackAll[g_index].p_index = -1;
			
			// read parent stack
			sprintf(data, "/proc/%s", filename->d_name);
			
			if (read_single_stack(data,g_index))
			{
				/*
				 * if we can access the stack of this
				 * process, continue to parse the child
				 * thread
				 */
				StackAll[g_index].pid = traced_process;
				g_index++;
				// read child stack
				//unwind to get the child stack,
				dump_sibling_thread_stack(atoi(filename->d_name), 
					g_index - 1);
			}

		}
	}
	closedir(dp);
	gettimeofday(&currentTime,0);
	read_all_stack_time = calc_timediff(&stillafstarttime, &currentTime);
	return 0;
}

void parse_kernel(int index)
{
	char *pKernel = StackAll[index].kernel_stack;
	printf("\nKernel Stack:\n");
	printf("%s", pKernel);
}

void *parse_all(void *arg)
{	
	pid_t traced_process;
	FILE *fp = NULL;
	mapinfo *milist = 0;
	int stack_depth = 0;
	int index = 0;
	int childcount = 0;
	struct timeval currentTime, stillafstarttime;

	gettimeofday(&stillafstarttime, 0);
	printf("\n\nparse_all, the stack whole size is %d\n", g_index);
	
	while (index <= g_index) {
			  
		char data[PATH_LENGTH];
		unsigned long data1[STACK_DEPTH];
		int iElfCount = 0;

		/* Clear stack pointer records */
		traced_process = StackAll[index].pid;
			
		if (traced_process < 1)
			break;

		printf("=========\n");
		if (StackAll[index].is_thread) {	
			sprintf(data, "/proc/%d/task/%d/comm",
				StackAll[index].ppid,
				StackAll[index].pid);
		}
		else {
			sprintf(data, "/proc/%d/comm",
				StackAll[index].pid);
		}

		fp = fopen(data,"r");
		if (!fp) {
			index++;
			printf("can't read %s\n", data);
			continue;
		}

		if(fgets(data, PATH_LENGTH, fp)) {
			strncpy(StackAll[index].p_name, data, strlen(data));
			if (StackAll[index].is_thread) {
				int pp_index = StackAll[index].p_index;

				printf("tid : %d, name: %s", StackAll[index].pid, data);
				printf("ppid: %d, name: %s\n", StackAll[index].ppid,
					StackAll[index].ppid > 0 ? StackAll[pp_index].p_name : " ");
			} 
			else {
				printf("pid : %d, name: %s", StackAll[index].pid, data);	
				int pp_index = StackAll[index].p_index;
				printf("ppid: %d, name: %s\n",
					StackAll[index].ppid,
					StackAll[index].ppid > 0 ? StackAll[pp_index].p_name : " ");			
				printf("child count is %d\n",StackAll[index].child_count);
			}

		}
		fclose(fp);
		fp = NULL;

		if (StackAll[index].is_thread)	{
			sprintf(data, "/proc/%d/task/%d/maps",
				StackAll[index].ppid,StackAll[index].pid);
		} else {
			sprintf(data, "/proc/%d/maps", StackAll[index].pid);
		}
				
		parse_kernel(index);

		//parse the map info of this process, create milist in here, need to free
		if(!StackAll[index].is_thread) {
			fp = fopen(data, "r");
			if(!fp)
				continue;

			while(fgets(data, PATH_LENGTH, fp)) {
				
				mapinfo *mi = parse_map_line_64(data);
				if(mi) {
					mi->next = milist;
					milist = mi;
					iElfCount ++;
				}
			}
			fclose(fp);
			fp = NULL;
		}


		if ((iElfCount < 1) && !StackAll[index].is_thread) {
			printf("can't access file /proc/%d/maps\n",traced_process);
			printf("\n\n\n");
			index++;
			continue;
		}

		//parse the elf of this process
		if (!StackAll[index].is_thread)
			parse_elf(milist, traced_process);
			
		// parse parent stack
		printf("\nUser Stack:\n");
		if(judge_64_OS()) 	
			unwind_backtrace_with_stack64(StackAll[index].ebp,
					StackAll[index].stack_depth,
					milist);
		else
			unwind_backtrace_with_stack(StackAll[index].ebp,
					StackAll[index].stack_depth,
					milist);

		printf("\n\n\n");

		//childcount ;
		if((StackAll[index].p_index >= 0)) {
			childcount++;
		}
		//free the milist
		if (StackAll[index].is_thread &&
			(childcount == StackAll[StackAll[index].p_index].child_count)) {
			while(milist) {
				mapinfo *next = milist->next;
				symbol_tables_free(milist->symbols);
				free(milist);
				milist = next;
			}
			childcount = 0;
		}
		index++;
	}

	g_index = 0;

	gettimeofday(&currentTime,0);
	printf("read all stack time: %ldms\n", read_all_stack_time);
	printf("parse_all time: %ldms\n",
		calc_timediff(&stillafstarttime, &currentTime));
	return NULL;
}

static int backtrace_android_whole()
{
	struct timeval currentTime, stillafstarttime;
	pid_t traced_process;
	long ins;
	FILE *fp;
	mapinfo *milist = 0;
	int stack_depth = 0;
	int index = 0;
	g_index = 0;

	gettimeofday(&stillafstarttime,0);
	memset(StackAll,0,sizeof(StackAll));

	read_all_stack();
	parse_all(NULL);

	gettimeofday(&currentTime,0);
	printf("Whole time: %ldms\n",
			calc_timediff(&stillafstarttime, &currentTime));
	return 0;
}
/*
////////////////////////////////////////////////////////////
void backtrace_parse_ANR_file( char *filename)
{
	pid_t traced_process;
	FILE *fp = NULL;
	mapinfo *milist = 0;
	unsigned int stack_depth = 0;
	int index = 0;
	int childcount = 0;
	struct timeval currentTime, stillafstarttime;
	char *str = NULL;
	int farther = false;
	FILE * fp_copy = NULL;
	int father_pid = -1;
	unsigned int data1[STACK_DEPTH];

	char data[PATH_LENGTH];
	char* result[STACK_DEPTH];
	unsigned long  userstack[STACK_DEPTH];
	unsigned int i = 0;

	sprintf(data,"%s_copy",filename);
	fp_copy  = fopen(data,"w");


	gettimeofday(&stillafstarttime, 0);


	fp  = fopen(filename,"r");
	while ( fp_copy && fp && fgets(data,PATH_LENGTH, fp) ) {
		int iElfCount = 0;
		if (strlen(data) < 2)
			continue;
		if (( str = strstr(data, "----- pid")) )  // got the farther task id,

			{
				str += 10;
				father_pid  = atoi(str);
				farther = true;
				fputs(data,fp_copy);
			}
		else {
			farther = false;
		}
		if (farther == true) //parse map info, skip write map info to new file. only read map info for farther task
		{    	
			while ( fgets(data,PATH_LENGTH,fp))  {
				if ( (str = strstr(data, "maps start")) )  // got the farther task id,
					{
						while(fgets(data, PATH_LENGTH, fp)) {
							mapinfo *mi = parse_map_line(data);
							if(mi) {
								mi->next = milist;
								milist = mi;
								iElfCount ++;
							}
							if ( strstr(data,"maps end"))
								goto f;
						}
					}

				else    {
					fputs(data,fp_copy);
				}
			}
		}
f:      if (farther == true)  {
		parse_elf32(milist, father_pid);
		farther = false;
	}
	while(fgets(data, PATH_LENGTH, fp)) {
		if ((str = strstr(data, "kernel stack"))) {
			while(fgets(data, PATH_LENGTH, fp)) {
				if ((str = strstr(data,"userspace"))) {
					while(fgets(data, PATH_LENGTH, fp)) {
						if (strlen(data) < 2) continue;
						if ((str = strstr(data, "userspace end"))) {
							break;
						}
						userstack[i++] = strtoul(data, NULL, 16);
						stack_depth++;
					}
					unwind_backtrace_with_stack_file(userstack,stack_depth,  milist, result);
					for ( i = 0; i < stack_depth; i++) {
						fwrite(result[i], strlen(result[i]), 1, fp_copy);

					}

					stack_depth = 0;
					break;

				}
				//write kernel stack infomation
				fwrite(data, strlen(data), 1, fp_copy);

			}
			//      fputs(data, fp_copy);
		}
		else  {
			if ((str = strstr(data, "----- end")))
				break;
			fputs(data,fp_copy);
		}
	}
	}
	return ;
}

void backtrace_parse_TOMBSTONE_file( char *filename)
{
	pid_t traced_process;
	FILE *fp = NULL;
	mapinfo *milist = 0;
	unsigned int stack_depth = 0;
	int index = 0;
	int childcount = 0;
	struct timeval currentTime, stillafstarttime;
	char *str = NULL;
	int farther = false;
	FILE * fp_copy = NULL;
	int father_pid = -1;
	unsigned int data1[STACK_DEPTH];

	char data[PATH_LENGTH];
	char* result[STACK_DEPTH];
	unsigned long  userstack[STACK_DEPTH];
	unsigned int i = 0;

	sprintf(data,"%s_copy",filename);
	fp_copy  = fopen(data,"w");


	gettimeofday(&stillafstarttime, 0);


	fp  = fopen(filename,"r");
	while ( fp_copy && fp && fgets(data,PATH_LENGTH, fp) ) {
		int iElfCount = 0;
		if (strlen(data) < 2)
			continue;
		if (( str = strstr(data, "----- pid")) )  // got the farther task id,

			{
				str += 10;
				father_pid  = atoi(str);
				farther = true;
				fputs(data,fp_copy);
			}
		else {
			farther = false;
		}
		if (farther == true) //parse map info, skip write map info to new file. only read map info for farther task
		{    	
			while ( fgets(data,PATH_LENGTH,fp))  {
				if ( (str = strstr(data, "maps start")) )  // got the farther task id,
					{
						while(fgets(data, PATH_LENGTH, fp)) {
							mapinfo *mi = parse_map_line(data);
							if(mi) {
								mi->next = milist;
								milist = mi;
								iElfCount ++;
							}
							if ( strstr(data,"maps end"))
								goto f;
						}
					}

				else    {
					fputs(data,fp_copy);
				}
			}
		}
f:      if (farther == true)  {
		parse_elf32(milist, father_pid);
		farther = false;
	}
	while(fgets(data, PATH_LENGTH, fp)) {
		if ((str = strstr(data, "kernel stack"))) {
			while(fgets(data, PATH_LENGTH, fp)) {
				if ((str = strstr(data,"userspace"))) {
					while(fgets(data, PATH_LENGTH, fp)) {
						if (strlen(data) < 2) continue;
						if ((str = strstr(data, "userspace end"))) {
							break;
						}
						userstack[i++] = strtoul(data, NULL, 16);
						stack_depth++;
					}
					unwind_backtrace_with_stack_file(userstack,stack_depth,  milist, result);
					for ( i = 0; i < stack_depth; i++) {
						fwrite(result[i], strlen(result[i]), 1, fp_copy);

					}

					stack_depth = 0;
					break;

				}
				//write kernel stack infomation
				fwrite(data, strlen(data), 1, fp_copy);

			}
			//      fputs(data, fp_copy);
		}
		else  {
			if ((str = strstr(data, "----- end")))
				break;
			fputs(data,fp_copy);
		}
	}
	}
	return ;
}


*/








////////////////////////////////////////////////////////////
int main(int argc, char **argv)
{
	int opt;
	char option[50];
	char path[256];
	pid_t pid = -1, p = -1;
	DIR * dp;
	struct dirent *filename;
	bool find = false;
	void* pp = NULL;
	unsigned int flag = 0;	

	//printf("xiaobing void size %d, long size %d \n", sizeof(pp), sizeof(long));
	//	bool is64OS = false;

	system("uname -a");
	is64OS  = judge_64_OS();
	/*	if (is64OS)
		{
		return 0;
		}
		*/
	while ((opt = getopt(argc, argv,"p:f:h")) != -1) {
		switch(opt) {
			case 'p':
				memcpy(option, optarg, 10);
				pid = strtoul(option, NULL, 10);
				break;

			case 'h':
				printf("Usage: parse_stack [-p process_id]\n");
				return 0;

			case 'f':
				memcpy(option, optarg, 50);
				flag = 1;
				break;
			case 'a':
				printf("parse all stack");
				pid = -1;

			default:
				printf("Invalid option\n");
				exit(1);
				break;	
		}	
	} 

	if (flag) {
		//	printf("filename: %s\n", option);
		//backtrace_parse_file(option);
		return 0;
	}
/*	if (is64OS)
	{

		dp = opendir("/proc/");
		if (!dp)
		{
			printf("open /proc directory error\n");
			return -1;
		}
		filename = readdir(dp);
		while (filename)
		{
			if(filename->d_name[0] > '0' &&  filename->d_name[0] <= '9')
			{		
				p = atoi(filename->d_name);
				if (p == pid)
					find = true;
			}
			filename = readdir(dp);
		}
		closedir(dp);

		if (pid != -1 && !find) {
			printf("The process %d doesn't exist.\n", pid);
			return -1;
		}

		if (find){
			g_index = 0;
			memset(StackAll,0,sizeof(StackAll));
			sprintf(path, "/proc/%d", pid);
			if (read_single_stack(path, g_index)) {
				StackAll[g_index].pid = pid;
				g_index++;
				dump_sibling_thread_stack(pid,  g_index-1);
				parse_all(NULL);
			}

		}

		if (pid < 0 && !find) {
			LOGV("###backtrace_android_whole\n");
			backtrace_android_whole();	
		}
		is64OS = false;
		return 0;
	}
*/
	dp = opendir("/proc/");
	if (!dp)
	{
		printf("open /proc directory error\n");
		return -1;
	}
	filename = readdir(dp);
	while (filename)
	{
		if(filename->d_name[0] > '0' &&  filename->d_name[0] <= '9')
		{		
			p = atoi(filename->d_name);
			if (p == pid)
				find = true;
		}
		filename = readdir(dp);
	}
	closedir(dp);

	if (pid != -1 && !find) {
		printf("The process %d doesn't exist.\n", pid);
		return -1;
	}

	if (find){
		g_index = 0;
		memset(StackAll,0,sizeof(StackAll));
		sprintf(path, "/proc/%d", pid);
		if (read_single_stack(path, g_index)) {
			StackAll[g_index].pid = pid;
			g_index++;
			dump_sibling_thread_stack(pid,  g_index-1);
			parse_all(NULL);
		}

	}

	if (pid < 0 && !find)
		backtrace_android_whole();
	return 0;
}

